# coding=utf-8
import numpy as np
import scipy.stats

from mixem.distribution.distribution import Distribution


class NormalDistribution(Distribution):
    """Univariate normal distribution with parameters (mu, sigma)."""

    def __init__(self, mu, sigma):
        self.mu = mu
        self.sigma = sigma

    def log_density(self, data):
        assert(len(data.shape) == 1), "Expect 1D data!"

        return - (data - self.mu) ** 2 / (2 * self.sigma ** 2) - np.log(self.sigma) - 0.5 * np.log(2 * np.pi)

    def estimate_parameters(self, data, weights):
        assert(len(data.shape) == 1), "Expect 1D data!"

        wsum = np.sum(weights)

        self.mu = np.sum(weights * data) / wsum
        self.sigma = np.sqrt(np.sum(weights * (data - self.mu) ** 2) / wsum)

    def __repr__(self):
        return "Norm[μ={mu:.4g}, σ={sigma:.4g}]".format(mu=self.mu, sigma=self.sigma)


class MultivariateNormalDistribution(Distribution):
    """Multivariate normal distribution with parameters (mu, Sigma)."""

    def __init__(self, mu, sigma):
        mu = np.array(mu)
        sigma = np.array(sigma)

        assert len(mu.shape) == 1, "Expect mu to be 1D vector!"
        assert len(sigma.shape) == 2, "Expect sigma to be 2D matrix!"

        assert sigma.shape[0] == sigma.shape[1], "Expect sigma to be a square matrix!"

        self.mu = mu
        self.sigma = sigma

    def log_density(self, data):
        return scipy.stats.multivariate_normal.logpdf(data, self.mu, self.sigma)

    def estimate_parameters(self, data, weights):
        self.mu = np.sum(data * weights[:, np.newaxis], axis=0) / np.sum(weights)

        center_x = data - self.mu[np.newaxis, :]

        # sigma = (np.diag(weights) @ center_x).T @ center_x / np.sum(weights)
        self.sigma = np.dot(
            np.dot(
                np.diag(weights),
                center_x
            ).T,
            center_x
        ) / np.sum(weights)

    def __repr__(self):
        po = np.get_printoptions()

        np.set_printoptions(precision=3)

        try:
            result = "MultiNorm[μ={mu}, σ={sigma}]".format(mu=self.mu, sigma=str(self.sigma).replace("\n", ","))
        finally:
            np.set_printoptions(**po)

        return result



# coding=utf-8
import numpy as np

from mixem.distribution.distribution import Distribution


class ExponentialDistribution(Distribution):
    """Exponential distribution with parameter (lambda)."""

    def __init__(self, lmbda):
        self.lmbda = lmbda

    def log_density(self, data):
        assert(len(data.shape) == 1), "Expect 1D data!"
        return np.log(self.lmbda) - self.lmbda * data

    def estimate_parameters(self, data, weights):
        assert(len(data.shape) == 1), "Expect 1D data!"
        self.lmbda = np.sum(weights) / np.sum(weights * data)

    def __repr__(self):
        return "Exp[λ={lmbda:.4g}]".format(lmbda=self.lmbda)



"""
Simple Expectation-Maximization fitting of mixtures of probability densities.
"""


import mixem.distribution
from mixem.progress import simple_progress
from mixem.em import em
from mixem.model import probability



import numpy as np


def probability(data, weights, distributions):
    """Compute the probability for data of the mixture density model given by weights and a list of distributions"""

    assert len(weights) == len(distributions), "Need matching number of weights and distributions!"
    data = np.array(data)

    n_dist = len(distributions)

    densities = np.empty((data.shape[0], n_dist))
    for d in range(n_dist):
        densities[:, d] = np.exp(distributions[d].log_density(data))

    return np.sum(weights[np.newaxis, :] * densities, axis=1)



import abc


class Distribution(object):
    """
    Base class for a mixEM probability distribution.

    To define your own new distribution, all methods of this class will have to be implemented.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def log_density(self, data):
        """Compute the log-probability density :math:`\log P(x|\phi)`

        :param data: The data :math:`x` to compute a probability density for. A :math:`(N \\times D)` :class:`numpy.ndarray` where N is the number of examples and D is the dimensionality of the data
        :type data: numpy.ndarray

        :returns: The log-probability for observing the data, given the probability distribution's parameters
        :rtype: float
        """
        raise NotImplementedError("Need to implement density calculation!")

    @abc.abstractmethod
    def estimate_parameters(self, data, weights):
        """Estimate the probabilities' parameters using weighted maximum-likelihood estimation and update parameters in-place.

        :param data: The data :math:`x` to estimate parameters for. A :math:`(N \\times D)` :class:`numpy.ndarray` where N is the number of examples and D is the dimensionality of the data
        :type data: numpy.ndarray

        :param weights: The weights :math:`\gamma` for individual data points. A N-element :class:`numpy.ndarray` where N is the number of examples.

        Choose those parameters :math:`\phi` that maximize the weighted log-likelihood function:

        .. math::
            ll_\gamma(x|\phi) = \sum_{n=1}^N \gamma_{n} \log [P(x|\phi)]

        Generally, this will involve differentiating the log-likelihood function for all parameters.
        You can set the derivative of the gradient to 0 and try to solve for the parameter to find
        a closed-form solution for the maximum-likelihood estimate or use a numerical optimizer to
        find the maximum-likelihood parameter.

        Once parameter estimates are found, update the attributes *in place*.
        """
        raise NotImplementedError("Need to implement parameter estimation!")

    @abc.abstractmethod
    def __repr__(self):
        """Create a string representation of the probability distribution"""
        raise NotImplementedError("Need to implement string representation!")



# coding=utf-8
import numpy as np

from mixem.distribution.distribution import Distribution


class GeometricDistribution(Distribution):
    """Geometric distribution with parameter (p)."""

    def __init__(self, p):
        self.p = p

    def log_density(self, data):
        assert(len(data.shape) == 1), "Expect 1D data!"
        return np.log(1 - self.p) * (data - 1) + np.log(self.p)

    def estimate_parameters(self, data, weights):
        assert(len(data.shape) == 1), "Expect 1D data!"

        self.p = np.sum(weights) / np.sum(data * weights)

    def __repr__(self):
        return "Geom[p={p:.4g}]".format(p=self.p)



import numpy as np

import mixem


def em(data, distributions, initial_weights=None, max_iterations=100, tol=1e-15, tol_iters=10, progress_callback=mixem.simple_progress):
    """Fit a mixture of probability distributions using the Expectation-Maximization (EM) algorithm.

    :param data: The data to fit the distributions for. Can be an array-like or a :class:`numpy.ndarray`
    :type data: numpy.ndarray

    :param distributions: The list of distributions to fit to the data.
    :type distributions: list of :class:`mixem.distribution.Distribution`

    :param initial_weights:  Inital weights for the distributions. Must be the same size as distributions. If None, will use uniform initial weights for all distributions.
    :type initial_weights: numpy.ndarray

    :param max_iterations:  The maximum number of iterations to compute for.
    :type max_iterations: int

    :param tol: The minimum relative increase in log-likelihood after tol_iters iterations
    :type tol: float

    :param tol_iters: The number of iterations to go back in comparing log-likelihood change
    :type tol_iters: int

    :param progress_callback: A function to call to report progress after every iteration.
    :type progress_callback: function or None

    :rtype: tuple (weights, distributitons, log_likelihood)
    """

    n_distr = len(distributions)
    n_data = data.shape[0]

    if initial_weights is not None:
        weight = np.array(initial_weights)
    else:
        weight = np.ones((n_distr,))

    last_ll = np.zeros((tol_iters, ))
    resp = np.empty((n_data, n_distr))
    log_density = np.empty((n_data, n_distr))

    iteration = 0
    while True:
        # E-step #######

        # compute responsibilities
        for d in range(n_distr):
            log_density[:, d] = distributions[d].log_density(data)

        # normalize responsibilities of distributions so they sum up to one for example
        resp = weight[np.newaxis, :] * np.exp(log_density)
        resp /= np.sum(resp, axis=1)[:, np.newaxis]

        log_likelihood = np.sum(resp * log_density)

        # M-step #######
        for d in range(n_distr):
            distributions[d].estimate_parameters(data, resp[:, d])

        weight = np.mean(resp, axis=0)

        if progress_callback:
            progress_callback(iteration, weight, distributions, log_likelihood)

        # Convergence check #######
        if np.isnan(log_likelihood):
            last_ll[0] = log_likelihood
            break

        if iteration >= tol_iters and (last_ll[-1] - log_likelihood) / last_ll[-1] <= tol:
            last_ll[0] = log_likelihood
            break

        if iteration >= max_iterations:
            last_ll[0] = log_likelihood
            break

        # store value of current iteration in last_ll[0]
        # and shift older values to the right
        last_ll[1:] = last_ll[:-1]
        last_ll[0] = log_likelihood

        iteration += 1

    return weight, distributions, last_ll[0]



# coding=utf-8


def simple_progress(iteration, weights, distributions, log_likelihood):
    """A simple default progress callback to use with mixem.em"""

    print("iteration {iteration:4d} (log-likelihood={log_likelihood:.5e}): p(x|Φ) = {formatted_distributions}".format(
        iteration=iteration,
        log_likelihood=log_likelihood,
        formatted_distributions=" + ".join("{w:.3g}*{d}".format(w=w, d=d) for w, d in zip(weights, distributions))
    ))
